# Takes lists of objects returned by the zkclient module, and
# consolidates the information for display.

import logging

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

logger = logging.getLogger('kafka.codec').addHandler(NullHandler())

import struct
import socket
from collections import namedtuple
from kafka.client import KafkaClient
from kafka.common import OffsetRequest

class ProcessorError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

PartitionState = namedtuple('PartitionState',
    [
        'broker',           # Broker host
        'topic',            # Topic on broker
        'partition',        # The partition
        'earliest',         # Earliest offset within partition on broker
        'latest',           # Current offset within partition on broker
        'depth',            # Depth of partition on broker.
        'spout',            # The Spout consuming this partition
        'current',          # Current offset for Spout
        'delta'             # Difference between latest and current
    ])
PartitionsSummary = namedtuple('PartitionsSummary',
    [
        'total_depth',      # Total queue depth.
        'total_delta',      # Total delta across all spout tasks.
        'num_partitions',   # Number of partitions.
        'num_brokers',      # Number of Kafka Brokers.
        'partitions'        # Tuple of PartitionStates
    ])

def process(spouts):
    '''
    Returns a named tuple of type PartitionsSummary.
    '''
    results = []
    total_depth = 0
    total_delta = 0
    brokers = []
    for s in spouts:
        for p in s.partitions:
            try:
                k = KafkaClient(p['broker']['host'], str(p['broker']['port']))
            except socket.gaierror, e:
                raise ProcessorError('Failed to contact Kafka broker %s (%s)' %
                                     (p['broker']['host'], str(e)))
            earliest_off = OffsetRequest(str(p['topic']), p['partition'], -2, 1)
            latest_off = OffsetRequest(str(p['topic']), p['partition'], -1, 1)
            earliest = k.send_offset_request([earliest_off])[0]
            latest = k.send_offset_request([latest_off])[0]
            current = p['offset']

            brokers.append(p['broker']['host'])
            total_depth = total_depth + (latest.offsets[0] - earliest.offsets[0])
            total_delta = total_delta + (latest.offsets[0] - current)

            results.append(PartitionState._make([
                p['broker']['host'],
                p['topic'],
                p['partition'],
                earliest.offsets[0],
                latest.offsets[0],
                latest.offsets[0] - earliest.offsets[0],
                s.id,
                current,
                latest.offsets[0] - current]))
    return PartitionsSummary(total_depth=total_depth,
                             total_delta=total_delta,
                             num_partitions=len(results),
                             num_brokers=len(set(brokers)),
                             partitions=tuple(results))



import simplejson as json
from collections import namedtuple

from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError

ZkKafkaBroker = namedtuple('ZkKafkaBroker', ['host', 'port'])
ZkKafkaSpout = namedtuple('ZkKafkaSpout', ['id', 'partitions'])
ZkKafkaTopic = namedtuple('ZkKafkaTopic', ['topic', 'broker', 'num_partitions'])

class ZkError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class ZkClient:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.client = KazooClient(hosts=':'.join([host, str(port)]))

    @classmethod
    def _zjoin(cls, e):
        return '/'.join(e)

    def brokers(self, broker_root='/brokers'):
        '''
        Returns a list of ZkKafkaBroker tuples, where each value is a
        ZkKafkaBroker.
        '''
        b = []
        id_root = self._zjoin([broker_root, 'ids'])

        self.client.start()
        try:
            for c in self.client.get_children(id_root):
                n = self.client.get(self._zjoin([id_root, c]))[0]
                b.append(ZkKafkaBroker(c, n.split(':')[1], n.split(':')[2]))
        except NoNodeError:
            raise ZkError('Broker nodes do not exist in Zookeeper')
        self.client.stop()
        return b

    def topics(self, broker_root='/brokers'):
        '''
        Returns a list of ZkKafkaTopic tuples, where each tuple represents
        a topic being stored in a broker.
        '''
        topics = []
        t_root = self._zjoin([broker_root, 'topics'])

        self.client.start()
        try:
            for t in self.client.get_children(t_root):
                for b in self.client.get_children(self._zjoin([t_root, t])):
                    n = self.client.get(self._zjoin([t_root, t, b]))[0]
                    topics.append(ZkKafkaTopic._make([t, b, n]))
        except NoNodeError:
            raise ZkError('Topic nodes do not exist in Zookeeper')
        self.client.stop()
        return topics

    def spouts(self, spout_root, topology):
        '''
        Returns a list of ZkKafkaSpout tuples, where each tuple represents
        a Storm Kafka Spout.
        '''
        s = []
        self.client.start()
        try:
            for c in self.client.get_children(spout_root):
                j = json.loads(self.client.get(self._zjoin([spout_root, c]))[0], encoding='utf-8')
                if j['topology']['name'] == topology:
                    s.append(ZkKafkaSpout._make([c, [j]]))
        except NoNodeError:
            raise ZkError('Kafka Spout nodes do not exist in Zookeeper')
        self.client.stop()
        return tuple(s)



#!/usr/bin/env python

import argparse
import sys
from prettytable import PrettyTable
import requests
import simplejson as json

from zkclient import ZkClient, ZkError
from processor import process, ProcessorError

def sizeof_fmt(num):
    for x in [' bytes','KB','MB','GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')

def null_fmt(num):
    return num

def display(summary, friendly=False):
    if friendly:
        fmt = sizeof_fmt
    else:
        fmt = null_fmt

    table = PrettyTable(['Broker', 'Topic', 'Partition', 'Earliest', 'Latest',
                        'Depth', 'Spout', 'Current', 'Delta'])
    table.align['broker'] = 'l'

    for p in summary.partitions:
        table.add_row([p.broker, p.topic, p.partition, p.earliest, p.latest,
                      fmt(p.depth), p.spout, p.current, fmt(p.delta)])
    print table.get_string(sortby='Broker')
    print
    print 'Number of brokers:       %d' % summary.num_brokers
    print 'Number of partitions:    %d' % summary.num_partitions
    print 'Total broker depth:      %s' % fmt(summary.total_depth)
    print 'Total delta:             %s' % fmt(summary.total_delta)

def true_or_false_option(option):
    if option == None:
        return False
    else:
        return True

def read_args():
    parser = argparse.ArgumentParser(
        description='Show complete state of Storm-Kafka consumers')
    parser.add_argument('--zserver', default='localhost',
        help='Zookeeper host (default: localhost)')
    parser.add_argument('--zport', type=int, default=2181,
        help='Zookeeper port (default: 2181)')
    parser.add_argument('--topology', type=str, required=True,
        help='Storm Topology')
    parser.add_argument('--spoutroot', type=str, required=True,
        help='Root path for Kafka Spout data in Zookeeper')
    parser.add_argument('--friendly', action='store_const', const=True,
                    help='Show friendlier data')
    parser.add_argument('--postjson', type=str,
                    help='endpoint to post json data to')
    return parser.parse_args()

def main():
    options = read_args()

    zc = ZkClient(options.zserver, options.zport)

    try:
        zk_data = process(zc.spouts(options.spoutroot, options.topology))
    except ZkError, e:
        print 'Failed to access Zookeeper: %s' % str(e)
        return 1
    except ProcessorError, e:
        print 'Failed to process: %s' % str(e)
        return 1
    else:
        display(zk_data, true_or_false_option(options.friendly))

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python

import setuptools

install_requires = [
    'PrettyTable==0.7.2',
    'kazoo==1.00',
    'simplejson',
    'argparse',
    'requests',
    'kafka-python'
]

setuptools.setup(
    name = 'stormkafkamon',
    version = '0.1.1',
    license = 'Apache',
    description = '''Monitor offsets of a storm kafka spout.''',
    author = "Philip O'Toole",
    author_email = 'philipomailbox-github@yahoo.com',
    url = 'https://github.com/vivekrao1985/stormkafkamon',
    platforms = 'any',
    packages = ['stormkafkamon'],
    zip_safe = True,
    verbose = False,
    install_requires = install_requires,
    dependency_links = ['https://github.com/mumrah/kafka-python/tarball/0.7#egg=kafka-python-0.7.2-0'],
    entry_points={
        'console_scripts': [
            'skmon = stormkafkamon.monitor:main'
        ]
    },
)
